
###########################################################################################
### internal functions taken from gamboostLSS 1.2-0

## helper functions
check <- function(what, what_char, names) {

  errormsg <- paste0(sQuote(what_char), " can be either a scalar, a (named) vector or a (named) list",
                     " of ", what_char, " values with same names as ",  sQuote("families"), "in ",
                     sQuote("boost_control"))

  if (is.list(what)) {
    if (is.null(names(what)) && length(what) == length(names))
      names(what) <- names
    if (!all(names(what) %in% names) ||
        length(unique(names(what))) != length(names))
      stop(errormsg)
    what <- what[names] ## sort in order of families
    what <- unlist(what)
  } else {
    if(length(what) != 1 && length(what) != length(names))
      stop(errormsg)
    if (length(what) == 1) {
      what <- rep(what, length(names))
      names(what) <- names
    } else {
      if (is.null(names(what)))
        names(what) <- names
      if (!all(names(what) %in% names))
        stop(errormsg)
      what <- what[names] ## sort in order of families
    }
  }

  return(what)
}


## helper function in a modified version based on mboost_2.2-3
## print trace of boosting iterations
do_trace <- function(current, mstart, risk,
                     linebreak = options("width")$width / 2, mstop = 1000) {
  current <- current - mstart
  if (current != mstop) {
    if ((current - 1) %/% linebreak == (current - 1) / linebreak) {
      mchr <- formatC(current + mstart, format = "d",
                      width = nchar(mstop) + 1, big.mark = "'")
      cat(paste("[", mchr, "] ",sep = ""))
    } else {
      if ((current %/% linebreak != current / linebreak)) {
        cat(".")
      } else {
        cat(" -- risk:", risk[current + mstart], "\n")
      }
    }
  } else {
    cat("\nFinal risk:", risk[current + mstart], "\n")
  }
}

## helper function copied from mboost_2.2-3
rescale_weights <- function(w) {
  if (max(abs(w - floor(w))) < sqrt(.Machine$double.eps))
    return(w)
  return(w / sum(w) * sum(w > 0))
}

## helper function copied from mboost_2.2-3
### check measurement scale of response for some losses
check_y_family <- function(y, family){
  ## <SB> check the response is in long format, otherwise is.numeric()
  ## as e.g. in StudentTMu()@check_y() gives an error
  family@check_y(as.vector(y))
  y
}



#################################################################################
#' Model-based Gradient Boosting for Functional GAMLSS
#'
#' Function for fitting GAMLSS (generalized additive models for location, scale and shape)
#' with functional data using component-wise gradient boosting, for details see
#' Brockhaus et al. (2015a).
#'
#' @param formula a symbolic description of the model to be fit.
#' If \code{formula} is a single formula, the same formula is used for all distribution parameters.
#' \code{formula} can also be a (named) list, where each list element corresponds to one distribution
#' parameter of the GAMLSS distribution. The names must be the same as in the \code{families}.
#' @param timeformula one-sided formula for the expansion over the index of the response.
#' For a functional response \eqn{Y_i(t)} typically \code{~bbs(t)} to obtain a smooth
#' expansion of the effects along \code{t}. In the limiting case that \eqn{Y_i} is a scalar response
#' use \code{~bols(1)}, which sets up a base-learner for the scalar 1.
#' Or you can use \code{timeformula=NULL}, then the scalar response is treated as scalar.
#' Analogously to \code{formula}, \code{timeformula} can either be a one-sided formula or
#' a named list of one-sided formulas.
#' @param data a data frame or list containing the variables in the model.
#' @param families an object of class \code{families}. It can be either one of the pre-defined distributions
#' that come along with the package \code{gamboostLSS} or a new distribution specified by the user
#' (see \code{\link[gamboostLSS]{Families}} for details).
#' Per default, the two-parametric \code{\link[gamboostLSS]{GaussianLSS}} family is used.
#' @param control  a list of parameters controlling the algorithm.
#' For more details see \code{\link[mboost]{boost_control}}.
#' @param weights does not work!
#' @param method fitting method, currently three methods are supported:
#' \code{"cycling"}, \code{"inner"} and \code{"outer"}.
#' The latter two require a one dimensional \code{mstop} value.
#' @param ... additional arguments passed to \code{\link[FDboost]{FDboost}},
#' including, \code{family} and \code{control}.
#'
#' @details For details on the theory of GAMLSS see Rigby and Stasinopoulos (2005).
#' \code{FDboostLSS} uses \code{FDboost} to fit the distibution parameters of a GAMLSS -
#' a functional boosting model is fitted for each parameter.
#' See \code{\link{FDboost}} for details on boosting functional regression models
#' as introduced by Brockhaus et al. (2015b).
#' See \code{\link[gamboostLSS]{mboostLSS}} for
#' details on boosting of GAMLSS for scalar variables as introduced by Mayr et al. (2012).
#'
#' @return an object of class \code{FDboostLSS} that inherits from \code{mboostLSS}.
#' The \code{FDboostLSS}-object is a named list containing one list entry per distribution parameter
#' and some attributes. The list is named like the parameters, e.g. mu and sigma,
#' if the parameters mu and sigma are modelled. Each list-element is an object of class \code{FDboost}.
#'
#' @author Sarah Brockhaus
#'
#' @seealso Note that \code{FDboostLSS} calls \code{\link{FDboost}} directly.
#'
#' @keywords models, nonlinear
#'
#' @references
#' Brockhaus, S. and Fuest, A. and Mayr, A. and Greven, S. (2015a):
#' Functional regression models for location, scale and shape applied to stock returns.
#' In: Friedl H. and Wagner H. (eds),
#' Proceedings of the 30th International Workshop on Statistical Modelling: 117-122.
#'
#' Brockhaus, S., Scheipl, F., Hothorn, T. and Greven, S. (2015b).
#' The functional linear array model. Statistical Modelling, 15(3), 279-300.
#'
#' Mayr, A., Fenske, N., Hofner, B., Kneib, T. and Schmid, M. (2012):
#' Generalized additive models for location, scale and shape for high-dimensional
#' data - a flexible approach based on boosting.
#' Journal of the Royal Statistical Society: Series C (Applied Statistics), 61(3), 403-427.
#'
#' Rigby, R. A. and D. M. Stasinopoulos (2005):
#' Generalized additive models for location, scale and shape (with discussion).
#' Journal of the Royal Statistical Society: Series C (Applied Statistics), 54(3), 507-554.
#'
#' @export
## function that calls FDboost for each distribution parameter

## Code based on non_cyclical fitting branch on GitHub by Janek Thomas.
## Latest commit 38634d2 on 11 Dec 2015
FDboostLSS <- function(formula, timeformula, data = list(), families = GaussianLSS(),
                       control = boost_control(), weights = NULL,
                       method = c("cycling", "inner", "outer"), ...){

  cl <- match.call()
  method <- match.arg(method)
  if(is.null(cl$families)) cl$families <- families

  ## warnings for functional response are irrelevant for scalar response
  if( !is.null(timeformula) && timeformula != ~bols(1) ){
    message("No smooth offsets over time are used, just global scalar offsets.")
    message("No integration weights are used to compute the loss for the functional response.")
  }

  fit <- FDboostLSS_fit(formula = formula, timeformula = timeformula,
                        data = data, families = families,
                        control = control, weights = weights, ...,
                        fun = FDboost, funchar = "FDboost", call = cl,
                        method = method)

  return(fit)
}

### work horse for fitting (glm/gam/m/black)boostLSS models
FDboostLSS_fit <- function(formula, timeformula,
                           data = list(), families = GaussianLSS(),
                           control = boost_control(), weights = NULL,
                           fun = FDboost, funchar = "FDboost", call = NULL,
                           method, ...){

  if(method == "cycling")
  iBoost_cycling <- function(niter) {

    start <- sapply(fit, mstop)
    mvals <- vector("list", length(niter))
    for (j in 1:length(niter)){
      mvals[[j]] <- rep(start[j] + niter[j], max(niter))
      if (niter[j] > 0)
        mvals[[j]][1:niter[j]] <- (start[j] + 1):(start[j] + niter[j])
    }

    ENV <- lapply(mods, function(j) environment(fit[[j]]$subset))

    for (i in 1:max(niter)){
      for (j in mods){
        ## update value of nuisance parameters
        ## use response(fitted()) as this is much quicker than fitted(, type = response)
        for (k in mods[-j])
          assign(names(fit)[k], families[[k]]@response(fitted(fit[[k]], toFDboost = FALSE)),
                 environment(get("ngradient", environment(fit[[j]]$subset))))
        ## update value of u, i.e. compute ngradient with new nuisance parameters

        ENV[[j]][["u"]] <- ENV[[j]][["ngradient"]](ENV[[j]][["y"]], ENV[[j]][["fit"]],
                                                   ENV[[j]][["weights"]])
        # same as:
        # evalq(u <- ngradient(y, fit, weights), environment(fit[[j]]$subset))

        ## update j-th component to "m-th" boosting step
        fit[[j]][mvals[[j]][i]]
      }
      if (trace){
        firstRun <- firstRun
        ## which is the current risk? rev() needed to get the last
        ## list element with maximum length
        whichRisk <- names(which.max(rev(lapply(lapply(fit, function(x) x$risk()), length))))
        do_trace(current = max(sapply(mvals, function(x) x[i])),
                 mstart = ifelse(firstRun, 0, max(start)),
                 mstop = ifelse(firstRun, max(niter) + 1, max(niter)),
                 risk = fit[[whichRisk]]$risk())
      }
    }
    return(TRUE)
  }

  if(method == "outer")
  iBoost_outer <- function(niter){

    # this is the case for boosting from the beginning
    if(is.null(attr(fit, "combined_risk")) | niter == 0){
      combined_risk <- vapply(fit, risk, numeric(1))
    }

    best <- which(names(fit) == tail(names(combined_risk), 1))



    ENV <- lapply(mods, function(j) environment(fit[[j]]$subset))

    for (i in seq_len(niter)){

      ## update value of nuisance parameters
      ## use response(fitted()) as this is much quicker than fitted(, type = response)
      for( k in mods[-best]){
        assign(names(fit)[best], families[[best]]@response(fitted(fit[[best]], toFDboost = FALSE)),
               environment(get("ngradient", environment(fit[[k]]$subset))))

        #evalq(u <- ngradient(y, fit, weights), ENV[[k]])
        ENV[[k]][["u"]] <- ENV[[k]][["ngradient"]](ENV[[k]][["y"]],
                                                     ENV[[k]][["fit"]],
                                                     ENV[[k]][["weights"]])
      }

      if (funchar == "glmboost") {
        lik_risks <- list()
        coefs <- list()

        for(i in mods){
            coefs[[i]] <- evalq(environment(get("fit1"))[["est"]](u)/
                                  environment(get("fit1"))[["sxtx"]], envir = ENV[[i]])

          all_fitted <- sweep(environment(ENV[[i]][["fit1"]])[["X"]], 2,
                              coefs[[i]] ,`*`)

          lik_risks[[i]] <- sapply(1:ncol(all_fitted), function(j)
            ENV[[i]][["triskfct"]](ENV[[i]][["y"]],
                                   ENV[[i]][["fit"]] + nu[i] * all_fitted[,j]))
        }

        # do a mboost step per hand
        best <- which.min(vapply(lik_risks, min,
                                 FUN.VALUE = numeric(1), ...))

        ENV[[best]]$all_fitted <- all_fitted
        ENV[[best]]$lik_risks <- lik_risks
        ENV[[best]]$best <- best
        ENV[[best]]$coefs <- coefs

        evalq({
          xs <- which.min(lik_risks[[best]])
          basses <- list(model = c(coef = coefs[[best]][xs],
                                   xselect = xs,
                                   p = length(coefs[[best]])),
                         fitted = function() {
                           return(coefs[[best]][xs] * environment(get("fit1"))[["X"]][, xs, drop = FALSE])
                         })
          class(basses) <- c("bm_cwlin", "bm_lin", "bm")
          fit <- fit + nu * basses$fitted()
          u <- get("ngradient")(get("y"), fit, weights)
          mrisk[(mstop + 1)] <- get("triskfct")(get("y"), fit)
          ens[[(mstop + 1)]] <- basses
          xselect[(mstop + 1)] <- xs
          nuisance[[(mstop +1)]] <- family@nuisance()
          mstop <- mstop + 1},
          envir <- ENV[[best]])

      }else{  # all other baselearner
        risks <- list()
        for( i in mods){
          risks[[i]] <- evalq({
            ss_new <- lapply(get("blfit", envir = environment(get("basefit"))),
                             function(x) x(u))
            sapply(ss_new, function(x) get("riskfct")(get("y"), fit + nu * x$fitted(), weights))
          }, envir = ENV[[i]])
        }

        best <- which.min(vapply(risks, min,
                                 FUN.VALUE = numeric(1)))

        evalq({
          ss <- lapply(get("blfit", envir = environment(get("basefit"))),
                       function(x) x(u))
          xselect[mstop + 1] <- which.min(sapply(ss, function(x)
            get("riskfct")(get("y"), fit + nu * x$fitted(), weights)))
          fit <- fit + nu * ss[[tail(xselect, 1)]]$fitted()
          u <- get("ngradient")(get("y"), fit, weights)
          mrisk[(mstop + 1)] <- get("triskfct")(get("y"), fit)
          ens[[(mstop + 1)]] <- ss[[tail(xselect, 1)]]
          nuisance[[(mstop + 1)]] <- family@nuisance()
          mstop <- mstop + 1},
          envir = ENV[[best]])

      }

      combined_risk[(length(combined_risk) + 1)] <- tail(risk(fit[[best]]), 1)
      names(combined_risk)[length(combined_risk)] <- names(fit)[best]

      if (trace){
        do_trace(current = length(combined_risk[combined_risk != 0]),
                 mstart = ifelse(firstRun, length(fit) + 1,
                                 length(combined_risk[combined_risk != 0])),
                 mstop = ifelse(firstRun, niter - length(fit), niter),
                 risk = combined_risk[combined_risk != 0])
      }
    }
    combined_risk  <<- combined_risk
    return(TRUE)
  }

  if(method == "inner" )
  iBoost_inner <- function(niter){

    # this is the case for boosting from the beginning
    if(is.null(attr(fit, "combined_risk")) | niter == 0){
      combined_risk <- vapply(fit, risk, numeric(1))
    }

    best <- which(names(fit) == tail(names(combined_risk), 1))

    ENV <- lapply(mods, function(j) environment(fit[[j]]$subset))

    for (i in seq_len(niter)){

      ## update value of nuisance parameters
      ## use response(fitted()) as this is much quicker than fitted(, type = response)
      for( k in mods[-best]){
        assign(names(fit)[best], families[[best]]@response(fitted(fit[[best]], toFDboost = FALSE)),
               environment(get("ngradient", environment(fit[[k]]$subset))))
      }

      risks <- numeric(length(fit))
      for(b  in 1:length(fit)){
        st <- mstop(fit[[b]])
        fit[[b]][st + 1]
          #risks[b] <- evalq({riskfct(y, fit, weights)}, envir = ENV[[b]])
          risks[b] <- ENV[[b]][["riskfct"]](ENV[[b]][["y"]],
                                            ENV[[b]][["fit"]],
                                            ENV[[b]][["weights"]])
        fit[[b]][st]

        ## fit[[b]][st] is not enough to reduce the model back to beginning,
        ## so all these values have to be reduced, so that they are calculated
        ## correctly the next time
        evalq({xselect <- xselect[1:mstop];
        mrisk <- mrisk[1:mstop];
        ens <- ens[1:mstop];
        nuisance <- nuisance[1:mstop]},
        environment(fit[[b]]$subset))
      }

      best <- which.min(risks)

      ## update value of u, i.e. compute ngradient with new nuisance parameters
      ENV[[best]][["u"]] <- ENV[[best]][["ngradient"]](ENV[[best]][["y"]],
                                                       ENV[[best]][["fit"]],
                                                       ENV[[best]][["weights"]])
      # same as:
      # evalq(u <- ngradient(y, fit, weights), environment(fit[[j]]$subset))

      ## update selected component by 1
      fit[[best]][mstop(fit[[best]]) + 1]

      combined_risk[(length(combined_risk) + 1)] <- tail(risk(fit[[best]]), 1)
      names(combined_risk)[length(combined_risk)] <- names(fit)[best]

      if (trace){
        do_trace(current = length(combined_risk[combined_risk != 0]),
                 mstart = ifelse(firstRun, length(fit) + 1,
                                 length(combined_risk[combined_risk != 0])),
                 mstop = ifelse(firstRun, niter - length(fit), niter),
                 risk = combined_risk[combined_risk != 0])
      }
      combined_risk  <<- combined_risk
    }
    return(TRUE)
  }

  if (length(families) == 0)
    stop(sQuote("families"), " not specified")

  if ("offset" %in% names(list(...)))
    stop("Do not use argument ", sQuote("offset"),
         ". Please specify offsets via families")
  ### Use mu in "families" to specify offset in mu-family etc.

  if (is.list(formula)){
    if (!all(names(formula) %in% names(families)) ||
        length(unique(names(formula))) != length(names(families)))
      stop(sQuote("formula"), " can be either a formula or a named list",
           " of formulas with same names as ",  sQuote("families"), ".")
    ynames <- sapply(formula, function(fm) as.character(fm[[2]]))
    if (length(unique(ynames)) > 1)
      warning("responses differ for the components")
    response <- lapply(formula, function(fm)
      eval(as.expression(fm[[2]]), envir = data,
           enclos = environment(fm)))
    #response <- eval(as.expression(formula[[1]][[2]]), envir = data,
    #                 enclos = environment(formula[[1]]))
  } else {
    response <- eval(as.expression(formula[[2]]), envir = data,
                     enclos = environment(formula))
    tmp <- vector("list", length = length(families))
    names(tmp) <- names(families)
    for (i in 1:length(tmp))
      tmp[[i]] <- formula
    formula <- tmp
  }

  if (is.list(timeformula)){
    if (!all(names(timeformula) %in% names(families)) ||
        length(unique(names(timeformula))) != length(names(families)))
      stop(sQuote("timeformula"), " can be either a one-sided formula or a named list",
           " of timeformulas with same names as ",  sQuote("families"), ".")
  } else {
    tmp <- vector("list", length = length(families))
    names(tmp) <- names(families)
    for (i in 1:length(tmp))
      tmp[i] <- list(timeformula)
    timeformula <- tmp
  }


  mstop <- mstoparg <- control$mstop
  control$mstop <- 1
  if (method == "cycling"){
    mstop <- check(mstop, "mstop", names(families))
  }else{
    #check mstop for inner and outer loss fitting methods
    if (length(mstop) != 1 | mstop %% 1 != 0 | mstop < length(families))
      stop(sQuote("mstop"), " has to be an integer larger than ",
           length(families))
  }

  nu <- control$nu
  nu <- check(nu, "nu", names(families))

  if (is.list(control$risk) || is.list(control$center) || is.list(control$trace))
    stop(sQuote("risk"),", ", sQuote("center"), " and ", sQuote("trace") ,
         " cannot be lists in ", sQuote("boost_control"))

  trace <- control$trace
  control$trace <- FALSE

  ### <FIXME> SB: is the dealing with the weights correct??
  w <- weights
  if (is.null(weights)){
    if (!is.list(response)) {
      weights <- rep.int(1, NROW(response))

      ### <SB> expand weights if the response is a matrix (functional response)
      if( !is.null(dim(response)) && !any(dim(response)==1) ){
        weights <- rep.int(weights, ncol(response))
      }

    } else {
      weights <- rep.int(1, NROW(response[[1]]))

      ### <SB> expand weights if the response is a matrix (functional response)
      if( !is.null(dim(response[[1]])) && !any(dim(response[[1]])==1) ){
        weights <- rep.int(weights, ncol(response[[1]]))
      }

    }
  } ## weights=rep(1, ncol(response[[j]])*nrow(response[[j]]) )

  weights <- rescale_weights(weights)

  fit <- vector("list", length = length(families))
  names(fit) <- names(families)

  mods <- 1:length(fit)

  offset <- vector("list", length = length(mods))
  names(offset) <- names(families)
  for (j in mods){
    if (!is.list(response)) {
      response <- check_y_family(response, families[[j]])
      offset[[j]] <- families[[j]]@offset(y = c(response), w = weights)
    } else {
      response[[j]] <- check_y_family(response[[j]], families[[j]])
      offset[[j]] <- families[[j]]@offset(y = c(response[[j]]), w = weights)
    }
    ## SB: e.g. for Families 'GaussianLSS' the offsets of mu and sigma are
    ## both written into both environments of the Family for mu and sigma
    ## for the computation of ngradient()
    for (k in mods){ ## <FIXME> SB: do this double-loop outside of this j-loop?
      for (l in mods){
        if (!is.null(offset[[l]]))
          assign(names(offset)[l], families[[l]]@response(offset[[l]]),
                 environment(families[[k]]@ngradient))
      }
    }
  }

  for (j in mods){
    ## update value of nuisance parameters in families
    for (k in mods[-j]){
      if (!is.null(fit[[k]])){
        ## families[[k]]@response(mboost:::fitted.mboost(fit[[k]]))
        ## <SB> set toFDboost=FALSE, to get predictions always in long format, never as matrix
        ##assign(names(fit)[k], fitted(fit[[k]], type = "response", toFDboost = FALSE),
        assign(names(fit)[k], families[[k]]@response(fitted(fit[[k]], toFDboost = FALSE)),
               environment(families[[j]]@ngradient))
      }
    }
    ## use appropriate nu for the model
    control$nu <- nu[[j]]
    ## <FIXME> Do we need to recompute ngradient?
    ## SB: look at environment within family
    ## ls.str(environment(families[[j]]@ngradient))
    fit[[j]] <- do.call(fun, list(formula[[names(families)[[j]]]],
                                  timeformula = timeformula[[names(families)[[j]]]], ## <SB> timeformula for FDboost
                                  data = data, family = families[[j]],
                                  offset = "scalar", ## <SB> fixme: always use scalar offset?
                                  control=control, weights = w,
                                  ...))
  }
  if (trace)
    do_trace(current = 1, mstart = 0,
             mstop = max(mstop),
             risk = fit[[length(fit)]]$risk())

  ## select the correct fitting method
  if (method == "cycling") {
    if (any(mstop > 1)){
      firstRun <- TRUE
      ## actually go for initial mstop iterations!
      tmp <- iBoost_cycling(mstop - 1)
    }
  } else {
    if (mstop >= length(fit)){
      firstRun <- TRUE
      combined_risk <- NULL
      tmp <- ifelse(method == "inner",
                    iBoost_inner(mstop - length(fit)),
                    iBoost_outer(mstop - length(fit)))
    }
  }

  firstRun <- FALSE

  class(fit) <- c(paste(funchar, "LSS", sep=""), "mboostLSS")
  if(method != "cycling"){
    class(fit) <- c("nc_FDboostLSS", "nc_mboostLSS", class(fit))
  }

  ### update to a new number of boosting iterations mstop
  ### i <= mstop means less iterations than current
  ### i >  mstop needs additional computations
  ### updates take place in THIS ENVIRONMENT,
  ### some models are CHANGED!
  if(method == "cycling") {
    attr(fit, "subset") <- function(i) {
      i <- check(i, "mstop", names(families))

      msf <- mstop(fit)
      niter <- i - msf
      if (all(niter == 0)) {
        ## make nothing happen with the model
        minStart <- max(msf)
      } else {
        minStart <- min(msf[niter != 0], i[niter != 0])
      }

      ## check if minStart bigger than mstop of parameters that are not touched
      #if (length(msf[niter == 0]) > 0 && minStart < min(msf[niter == 0]))
      #minStart <- min(msf[niter == 0])

      ## reduce models first (when necessary)
      if (any(msf > minStart)){
        cf <- class(fit)
        class(fit) <- "list" ## needed to use [] operator for lists

        #cat("processed parameters: ", paste(names(fit[msf > minStart]),
        #                                    collapse = ", "), "\n")

        lapply(fit[msf > minStart],
               function(obj) obj$subset(minStart))

        ## remove additional boosting iterations from environments
        lapply(fit[msf > minStart], function(obj){
          evalq({xselect <- xselect[1:mstop];
          mrisk <- mrisk[1:mstop];
          ens <- ens[1:mstop];
          nuisance <- nuisance[1:mstop]},
          environment(obj$subset))
        })

        class(fit) <- cf
        if (trace)
          cat("Model first reduced to mstop = ", minStart, ".\n",
              "Now continue ...\n", sep ="")
      }

      ## now increase models (when necessary)
      if (any(i > minStart)){
        ## set negative values to 0
        ## (only applicable if some parameters do not need to be touched
        inc <- ifelse(i - minStart > 0, i - minStart, 0)
        tmp <- iBoost_cycling(inc)
      }

      mstop <<- i
    }
  } else {
    attr(fit, "subset") <- function(i) {
      if(i < length(fit)){
        warning(paste("Minimal number of iterations:", length(fit),
                      "(at least one iteration for each distribution parameter)"))
        i <- length(fit)
      }
      msf <- sum(mstop(fit))
      niter <- i - msf

      ## check if minStart bigger than mstop of parameters that are not touched
      #if (length(msf[niter == 0]) > 0 && minStart < min(msf[niter == 0]))
      #minStart <- min(msf[niter == 0])

      ## reduce models first (when necessary)
      if (niter < 0 ){
        cf <- class(fit)
        class(fit) <- "list" ## needed to use [] operator for lists

        #cat("processed parameters: ", paste(names(fit[msf > minStart]),
        #                                    collapse = ", "), "\n")

        # reduce the combined risk values
        combined_risk <<- attr(fit, "combined_risk")()[1:i]
        new_stop_value <- table(names(attr(fit, "combined_risk")()))


        for(o in names(new_stop_value)){
          fit[[o]]$subset(new_stop_value[o])
        }

        ## remove additional boosting iterations from environments
        lapply(fit, function(obj){
          evalq({xselect <- xselect[1:mstop];
          mrisk <- mrisk[1:mstop];
          ens <- ens[1:mstop];
          nuisance <- nuisance[1:mstop]},
          environment(obj$subset))
        })

        # update ALL nuisance parameters to last update
        ENV <- lapply(mods, function(j) environment(fit[[j]]$subset))
        for(j in names(new_stop_value)){
          for( k in setdiff(names(new_stop_value), j)){
            assign(k, families[[k]]@response(fitted(fit[[k]], toFDboost = FALSE)),
                   environment(get("ngradient", environment(fit[[j]]$subset))))
          }
        }

        for(k in mods){
          # evalq(u <- ngradient(y, fit, weights), ENV[[k]])
		  evalq(u <- get("ngradient")(get("y"), fit, weights), ENV[[k]])
        }


        class(fit) <- cf
      }

      ## now increase models (when necessary)
      else if (niter > 0){
        if (method == "inner") {
          tmp <- iBoost_inner(niter)
        } else {
          tmp <- iBoost_outer(niter)
        }

      }

      mstop <<- i
    }
  }

  ## make call in submodels nicer:
  cl <- call
  cl[[1]] <- as.name(gsub("LSS", "", cl[[1]]))
  names(cl)[names(cl) == "families"] <- "family"
  for (i in 1:length(fit)) {
    fit[[i]]$call <- cl
    ## <FIXME> This is not really nice
    fit[[i]]$call$family <- families[[i]]
  }

  attr(fit, "(weights)") <- weights  ## attach weights used for fitting

  ## update to new weights; just a fresh start
  attr(fit, "update") <- function(weights = NULL, oobweights = NULL,
                                  risk = NULL, trace = NULL, mstop = NULL) {
    if (is.null(mstop)) {
      control$mstop <- mstoparg
    } else {
      control$mstop <- mstop
    }
    if (!is.null(risk))
      control$risk <- risk
    if (!is.null(trace))
      control$trace <- trace
    ## re-use user specified offset only
    ## (since it depends on weights otherwise)
    ## this is achieved via a re-evaluation of the families argument
    FDboostLSS_fit(formula = formula, timeformula = timeformula,
                   data = data,
                   families = eval(call[["families"]]), weights = weights,
                   control = control, fun = fun, funchar = funchar,
                   call = call, oobweights = oobweights,
                   method = method)
  }
  attr(fit, "control") <- control
  attr(fit, "call") <- call
  attr(fit, "data") <- data
  attr(fit, "families") <- families
  if(method != "cycling")
    attr(fit, "combined_risk") <- function() combined_risk

  return(fit)
}


#################################################################################
#' Extract selected base-learners
#'
#' Extract selected base-learners, see also \code{selected.mboostLSS}.
#'
#' @param object an object of class 'FDboostLSS'
#'
#' @return list with an entry for each distribution parameter
#' giving the numbers of the base-learners that were selected in each
#' boosting-iteration.
#'
#' @export
selected.FDboost <- function(object, ...)
  object$xselect()


#################################################################################
#' Transform mboost model to FDboost model
#'
#' Experimental function. Prone to changes!
#' Transform 'mboost'-object to 'FDboost'-object, to use methods of class 'FDboost',
#' especially \code{coef.FDboost} and \code{plot.FDboost}.
#'
#' @param object an object of class 'mboost'
#' @param data the data that was used to fit the model
#' @param time the name of the time variable, for scalar response use 'ONEtime'
#' @param response the name of the response variable
#'
#' @return an object of class \code{FDboost}.
#'
#' @export
toFDboost <- function(object, data, time="time", response="y"){

  ## add the necessary extra information
  if(time == "ONEtime"){ # scalar response
    object$ydim <- c(length(object$response), 1)
    object$yind <- 1
    attr(object$yind, "nameyind") <- "ONEtime"
    object$timeformula <- "~bols(ONEtime, intercept = FALSE)"
  }else{ # functional response
    object$ydim <- dim(data[[ response ]] ) # dim(response)
    if(is.null(object$ydim)) object$ydim <- length(object$response)
    object$yind <- data[[time]]
    attr(object$yind, "nameyind") <- time
  }

  object$predictOffset <- function(time) return(rep(0, l = length(time)))
  object$formulaMboost <- object$call$formula

  class(object) <- c("FDboost", class(object))
  # if(!is.null(id)) class(object) <- c("FDboostLong", class(object))
  if(time == "ONEtime") class(object) <- c("FDboostScalar", class(object))

  object
}



