# plot functional data (rows=obs, columns=argvals) with lines
matlplot <- function(x, y, lty=1, col=rgb(0,0,0,.1), lwd=.1, ...){
  if (missing(x)) {
    if (missing(y))
      stop("must specify at least one of 'x' and 'y'")
    else x <- 1L:NROW(y)
  }
  else if (missing(y)) {
    y <- x
    x <- 1L:NROW(y)
  }
  matplot(x,y, type="l", lty=lty, col=col, lwd=lwd, ...)
}


#-------------------------------------------------------------------------------
# fit pffr model and save various (graphical) summaries and model diagnostics
runModel <- function(name, frml, data=dti, yind = segm.t,
  bs.yindex=list(bs="ps", k=25, m=c(2,1)), control=gam.control(trace=FALSE,
    nthreads=8),
  ...) {
  start <- Sys.time()
  assign(name, pffr(frml, bs.yindex = bs.yindex,  yind = yind, data=data,
    control=control, na.action=na.exclude, ...))
  elapsed <- difftime(Sys.time(), start)
  try(saveRDS(get(name),
    file=paste0("models/",name,".rds")))

  concpartial <- try(concurvity(get(name), full=FALSE))
  concfull <- try(concurvity(get(name), full=TRUE))
  try(saveRDS(list(partial = concpartial, full=concfull),
    file=paste0("models/",name,"-concurvity.rds")))

  sumtxt <- paste0("models/", name,"-summary.txt")
  try(capture.output(summary(get(name)), file = sumtxt))
  try(capture.output(print(c("AIC:", AIC(get(name)))), file=sumtxt, append=TRUE))
  try(capture.output(print(c("BIC:", BIC(get(name)))), file=sumtxt, append=TRUE))
  try(capture.output(print(elapsed), file =  sumtxt, append=TRUE))
  try(capture.output(print(sessionInfo()), file =  sumtxt, append=TRUE))
  try(capture.output(cat("\n ################# concurvity \n"), file =  sumtxt, append=TRUE))
  try(capture.output(print(concfull), file =  sumtxt, append=TRUE))
  try(capture.output(cat("\n ################# gam.check \n"), file =  sumtxt, append=TRUE))
  m_as_gam <- get(name)
  class(m_as_gam) <- class(m_as_gam)[-1]
  try(capture.output(gam.check(m_as_gam), file =  sumtxt, append=TRUE))
  pdf(paste0("models/", name,".pdf"))
  try(do.call(plot, list(x=get(name), pers=TRUE, ticktype="detailed")))
  if(grepl("ffpc", name)){
    try(ffpcplot(get(name)))
  }
  try(qq.gam(m_as_gam))
  try(image(concpartial$worst, breaks = seq(0, 1, l=13)))
  dev.off()
  try(assign(paste0("terms.",name),
    do.call(predict, list(object=get(name), type="terms"))))
  try(saveRDS(get(paste0("terms.",name)),
    file=paste0("models/",name,"-terms.rds")))
  try(assign(paste0("cm.", name), coef(get(name))))
  try(saveRDS(get(paste0("cm.",name)),
    file=paste0("models/",name,"-cm.rds")))
  try(assign(paste0("fit.", name), Reduce("+", get(paste0("terms.",name)))))
  try(saveRDS(get(paste0("fit.",name)),
    file=paste0("models/",name,"-fit.rds")))

  invisible(get(name))
}

#-------------------------------------------------------------------------------
# plot functional intercepts for treatment & control groups
plot_ms <- function(cm, trans=function(x) x, yylim=NULL, col=c("darkblue", "gold")){
  require(scales)
  pffr_ms = cm$smterms[["case1(yind)"]]
  pffr_ct = cm$smterms[["case0(yind)"]]

  ## contrast MS vs healthy
  ms <- data.frame(
    ind = pffr_ms$x,
    value =  pffr_ms$value + cm$pterms["case1", "value"],
    se = sqrt(pffr_ms$se^2 + cm$pterms["case1", "se"]^2)
    )
  ms$upper <- trans(ms$value + 2* ms$se)
  ms$lower <- trans(ms$value - 2* ms$se)
  ms$value <- trans(ms$value)

  ct <- data.frame(
    ind = pffr_ct$x,
    value =  pffr_ct$value + cm$pterms["case0", "value"],
    se = sqrt(pffr_ct$se^2 + cm$pterms["case0", "se"]^2)
  )
  ct$upper <- trans(ct$value + 2* ct$se)
  ct$lower <- trans(ct$value - 2* ct$se)
  ct$value <- trans(ct$value)

  par(mar=par()$mar-c(0,0,1,0))

  if(is.null(yylim)) {
    yylim <- c(.97,1.03)*range(c(ms$upper, ms$lower, ct$upper, ct$lower))
  }
  plot(ms$value, type="l", col=col[1], ylim=yylim, xlab="CCA tract location $t$",
    ylab="Mean FA-CCA", axes=F, cex.lab=1.3, lwd=2, lty=3)
  polygon(c(1:nrow(ms), nrow(ms):1), c(ms[,"upper"], rev(ms[,"lower"])),
    border = NA, col=scales::alpha(col[1], .2))
  lines(ct$value, col=col[2], lwd=2, lty=1)
  polygon(c(1:nrow(ct), nrow(ct):1), c(ct[,"upper"], rev(ct[,"lower"])),
    border = NA, col=scales::alpha(col[2], .2))
  axis(1,at=round(seq(0, max(ms$ind), len=5)), cex.axis=1.3)
  axis(2,at=round(seq(yylim[1], yylim[2], len=4), 2), cex.axis=1.3)
  legend("top", lty=c(3, 1), col=col, legend=c("MS", "control"),
    bty = "n", cex=1.3)
}


